@@ -1,6 +1,6 @@
 /* 
    HTTP Request Handling
-   Copyright (C) 1999-2002, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2008, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -25,19 +25,16 @@
 #ifndef NE_PRIVATE_H
 #define NE_PRIVATE_H
 
-#ifdef NEON_SSL
-#include <openssl/ssl.h>
-#endif
-
 #include "ne_request.h"
 #include "ne_socket.h"
+#include "ne_ssl.h"
 
 struct host_info {
     char *hostname;
     unsigned int port;
     ne_sock_addr *address; /* if non-NULL, result of resolving 'hostname'. */
     /* current network address obtained from 'address' being used. */
-    ne_inet_addr *current;
+    const ne_inet_addr *current;
     char *hostport; /* URI hostport segment */
 };
 
@@ -57,9 +54,11 @@ struct hook {
 struct ne_session_s {
     /* Connection information */
     ne_socket *socket;
+
+    /* non-zero if connection has been established. */
+    int connected;
     
-    /* non-zero if socket if socket has persisted beyond one
-     * request. */
+    /* non-zero if connection has persisted beyond one request. */
     int persisted;
 
     int is_http11; /* >0 if connected server is known to be
@@ -68,24 +67,19 @@ struct ne_session_s {
     char *scheme;
     struct host_info server, proxy;
 
-    /* Connection states:
-     *   0:  Not connected at all.
-     *   1:  We have a TCP connection to the next-hop server.
-     *   2:  We have a negotiated an SSL connection over the proxy's 
-     *       TCP tunnel.
-     *
-     * Note, 1 is all we need if we don't have a proxy server, or
-     * if we do have a proxy server and we're not using SSL.
-     */
-    unsigned int connected:2;
+    /* application-provided address list */
+    const ne_inet_addr **addrlist;
+    size_t numaddrs, curaddr;
+
+    /* Local address to which sockets should be bound. */
+    const ne_inet_addr *local_addr;
 
     /* Settings */
-    unsigned int use_proxy:1; /* do we have a proxy server? */
-    unsigned int no_persist:1; /* set to disable persistent connections */
-    unsigned int use_ssl:1; /* whether a secure connection is required */
-    unsigned int in_connect:1; /* doing a proxy CONNECT */
+    int use_proxy; /* do we have a proxy server? */
+    int use_ssl; /* whether a secure connection is required */
+    int in_connect; /* doing a proxy CONNECT */
 
-    int expect100_works; /* known state of 100-continue support */
+    int flags[NE_SESSFLAG_LAST];
 
     ne_progress progress_cb;
     void *progress_ud;
@@ -93,20 +87,21 @@ struct ne_session_s {
     ne_notify_status notify_cb;
     void *notify_ud;
 
-    int rdtimeout; /* read timeout. */
+    int rdtimeout, cotimeout; /* read, connect timeouts. */
 
-    struct hook *create_req_hooks, *pre_send_hooks, *post_send_hooks;
-    struct hook *destroy_req_hooks, *destroy_sess_hooks, *private;
+    struct hook *create_req_hooks, *pre_send_hooks, *post_send_hooks,
+        *post_headers_hooks, *destroy_req_hooks, *destroy_sess_hooks, 
+        *close_conn_hooks, *private;
 
-    char *user_agent; /* full User-Agent string */
+    char *user_agent; /* full User-Agent: header field */
 
-#ifdef NEON_SSL
-    SSL_CTX *ssl_context;
-    X509 *server_cert;
-    SSL_SESSION *ssl_sess;
-    /* client cert */
-    EVP_PKEY *client_key;
-    X509 *client_cert;
+#ifdef NE_HAVE_SSL
+    ne_ssl_client_cert *client_cert;
+    ne_ssl_certificate *server_cert;
+    ne_ssl_context *ssl_context;
+    int ssl_cc_requested; /* set to non-zero if a client cert was
+                           * requested during initial handshake, but
+                           * none could be provided. */
 #endif
 
     /* Server cert verification callback: */
@@ -115,22 +110,26 @@ struct ne_session_s {
     /* Client cert provider callback: */
     ne_ssl_provide_fn ssl_provide_fn;
     void *ssl_provide_ud;
-    /* Client cert key password private key provider callback: */
-    ne_ssl_keypw_fn ssl_keypw_fn;
-    void *ssl_keypw_ud;
+
+    ne_session_status_info status;
 
     /* Error string */
-    char error[BUFSIZ];
+    char error[512];
 };
 
+/* Pushes block of 'count' bytes at 'buf'. Returns non-zero on
+ * error. */
 typedef int (*ne_push_fn)(void *userdata, const char *buf, size_t count);
 
-/* Pulls the request body for the given request, passing blocks to the
- * given callback.
- */
-int ne_pull_request_body(ne_request *req, ne_push_fn fn, void *ud);
-
 /* Do the SSL negotiation. */
-int ne_negotiate_ssl(ne_request *req);
+int ne__negotiate_ssl(ne_session *sess);
+
+/* Set the session error appropriate for SSL verification failures. */
+void ne__ssl_set_verify_err(ne_session *sess, int failures);
+
+/* Return non-zero if hostname from certificate (cn) matches hostname
+ * used for session (hostname); follows RFC2818 logic.  cn is modified
+ * in-place. */
+int ne__ssl_match_hostname(char *cn, const char *hostname);
 
 #endif /* HTTP_PRIVATE_H */
